/*
Realisiert mit AT-Mega88, 16MHz

Ausgnge:
PD1     TX zum LCD


Eingnge:
PC0     Spannung
PC1     Strom
PC2     Offset
PB0     Nulldurchgang
PD0     Offsetmessung
PD2     Schalterstellung 400mA
PD3     Schalterstellung 100mA
PD4     Schalterstellung  25mA


==========================================
T0 - 8Bit:   Messung von Verweildauer ausserhalb des ADC-Interrupts
T1 - 16Bit:  Phasenwinkelmessung
T2 - 8bit:   Zeitbasis fr Delay-Funktion

Messung:
  Es werden 4 Vollwellen lang die Spannungswerte mit 8 Bit aufgezeichnet 
  Anschlieend wird eine Vollwelle ausgesetzt.
  Dann wird 4 Vollwellen lang der Stromverlauf mit 10 Bit gemessen und
  mit den aufgezeichneten Spannungswerten multipliziert.


Normierung Spannung:
  Theorie:
    230 Veff sind 325 Vmax
    Mit Spannungsteiler 400k/2K7 gibt das 2.18V Spitzenspannung
  
    Mit 5V Uref und 1024 Schritten des A/D Wandlers:
    1 Schritt sind 5/1024 V
    2.18V = 446 Schritte
    446/325=1.372
  
    Spannung: Schritte / 1.372

Normierung Strom:
  400mA geben 2.5V
  1 Schritt sind 5/1024 V
  400mA = 512 Schritte
  512/400=1.28
  
  Strom: Schritte / 1.28

  XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  Praxis:

    219V Netzspannung gemessen. Gibt 310V Scheitelspannung.
      1,04V -> 310V 
    ADC zeigt 512+212.
      212Schritte -> 310V
      U = Schritte * 1,462
  XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX


!!! ACHTUNG !!!
Fr Floating-Point Ausgabe folgende Libraries mit einbinden:
  libm.a
  libprintf_flt.a

und als Custom-Option fr den Linker
  -Wl,-u,vfprintf
eintragen.

siehe:
http://winavr.scienceprog.com/avr-gcc-tutorial/using-sprintf--function-for-float-numbers-in-avr-gcc.html


*/
/* 
Der Compiler erzeugt automatisch Code, welcher alle Konstanten beim Programmstart ins RAM kopiert.
Da RAM knapp ist - und ich nicht besonders schnell auf alle konstanten Texte zugreifen muss, bleiben sie im
Programmspeicher (ROM). Das verklickert man dem Compiler mit dem PROGMEM Makro aus <avr/pgmspace.h>.
Auf die Daten kann man dann nur Byte fr Byte mit der Funktion pgm_read_byte(adresse) zugreifen.
Um das etwas bequemer zu machen dient die Funktion void loadText(char *src), welche einen Text aus dem 
ROM ins RAM lcdGlob.message kopiert.
*/
#include "typedefs.h"
#include <avr/pgmspace.h>       // definitions for keeping constants in program memory



#define F_CPU 16000000L	/* CPU clock in Hertz */

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
//#include <stdio.h> //sprintf
#include <math.h>  //sqrt

typedef unsigned char byte;

#define U_KALIBRIER 1462L
#define I_KALIBRIER 1552L  
#define PHI_KALIBRIER 3466L



const U08 txt01[] PROGMEM = "|o|*|u|*|2     <- weiter drehen ->";
const U08 txt02[] PROGMEM = "|u|*|o|*|2I-Offset=123";
const U08 txt03[] PROGMEM = "|o|*|u|*|1Wirkleistungsmessgeraet";
const U08 txt04[] PROGMEM = "|3D. Schroeder 01/2009";
const U08 txt05[] PROGMEM = "|1|*P=123.567W|2Ueff=123V Ieff=123.5mA";
const U08 txt06[] PROGMEM = "|3|*Bereich ";
const U08 txt07[] PROGMEM = " 25  100 [400] mA";
const U08 txt08[] PROGMEM = " 25 [100] 400  mA";
const U08 txt09[] PROGMEM = "[25] 100  400  mA";
const U08 txt10[] PROGMEM = "|G4HUeberlast";
const U08 txt11[] PROGMEM = "|4Ni=123 Nu=123 Ioffs=123";
const U08 txt12[] PROGMEM = "|G1SNnull=1";
const U08 txt13[] PROGMEM = "|4Phi=+123 +456 +789 +123";
//const U08 txt13[] PROGMEM = "";


//#define STUETZSTELLEN 385//384 werden gebraucht
#define STUETZSTELLEN 800//384 werden gebraucht
//Der Einfachheit halber: viele globale Variablen

//Wir haben insgesamt 1024 Bytes SRAM -  Platz, um alle Variablen statisch zu halten
typedef struct {
  int offsetU;
  int offsetI;
  unsigned int countU;
  unsigned int countI;
//  unsigned int countPhi;
  long sumSpannung;
  long sumStrom;
  long sumWirk;
//  long winkel;
//  long sumWinkel;
//  byte modus;
  volatile byte messen;
//  volatile byte verweildauer;
//  volatile byte verweildauerMax;
//  volatile byte verweildauerMin;
  volatile unsigned int ms;
  volatile byte finished;
  int LemWindungen;
    
//  long Umax;
//  long Imax;
  signed char istwerteU[STUETZSTELLEN]; //Bei 50Hz und 9600 Samples/s kriege ich 192 Messwerte je Vollwelle
  volatile byte spannungsmessung;
  volatile byte trigger;
  volatile S16 nulldurchgangszeitpunkt[20];
  volatile S16 phi[8];
  volatile U08 nulldurchgaenge; 
  volatile U08 error;
  U08 message[50];

}Tglob;

//Alle globalen Var in einer Struct deklarieren
Tglob g;


//--------------------------------------------------------
//--------------------------------------------------------
//--------------------------------------------------------
//--------------------------------------------------------
void loadText(const U08 *src){
  U08 *dst;
  dst = g.message;
  while ((*dst++ = pgm_read_byte(src++)))
  ;
}

//--------------------------------------------------------
// ms - Timer
ISR(TIMER2_COMPA_vect){ 
  g.ms--;
  if (0==g.ms){  //Bewertung muss innerhalb der IRQ-Routine erfolgen,
                 //da Bewertung der 16Bit Variable sonst unterbrochen 
                 //werden kann!
    g.finished = 1;
  }
}
//--------------------------------------------------------
//ADC-Ergebnis fertig
ISR(ADC_vect){
  //ADC liefert mit 9,6KHz einen Messwert
  static int messwert;
  static long istwert;
  static int u;
  U16 zeit;

  messwert = ADC;
  zeit = TCNT1;
  if (g.messen){
    ADCSRA |= _BV(ADSC); //Start next conversion
    
    if (g.spannungsmessung) { 
      istwert = (messwert - g.offsetU) ;
      if ((istwert >= 0) && g.trigger){
        g.nulldurchgangszeitpunkt[g.nulldurchgaenge++] = zeit;   
        g.trigger=0;
      }

      if (g.countU >= STUETZSTELLEN){
        g.countU = STUETZSTELLEN - 1;
      }
      g.istwerteU[g.countU++] = istwert/4;
      g.sumSpannung += istwert * istwert;
      if (istwert < -200L) g.trigger=1;
        
    }else{
      istwert = (messwert - g.offsetI);
      if ((istwert >= 0)  && g.trigger){
        if (g.nulldurchgaenge < 4){
          g.phi[g.nulldurchgaenge] = zeit - g.nulldurchgangszeitpunkt[g.nulldurchgaenge];   
        }
        g.trigger=0;
        if (g.nulldurchgaenge< 99){
          g.nulldurchgaenge++;
        }
        
      }
      g.sumStrom += istwert * istwert;
      if (g.countI >= STUETZSTELLEN){
        g.countI = STUETZSTELLEN - 1;
        g.error|=1;
      }
      u = g.istwerteU[g.countI++];
      g.sumWirk += u*4 * istwert;
      if (istwert < -10L) g.trigger=1;
    }
  }else{
    g.trigger=0;
  }
}  
//--------------------------------------------------------
//--------------------------------------------------------
//--------------------------------------------------------
void offsetUMessung(void){
  byte i;
  int sum;
  ADMUX = _BV(REFS0)|2; //Internal 5V Ref; Right adjust; MUX=Ch2 (Offset-Kanal)
  ADCSRA = _BV(ADEN)| //Enable
          _BV(ADSC) | //Start conversion
          _BV(ADATE)| //Auto trigger (free running )
          _BV(ADPS2)| //Prescaler=128  
          _BV(ADPS1)|
          _BV(ADPS0);
  ADCSRB = 0; //free running

  ADCSRA |= _BV(ADIF); //Clear ADIF by writing 1 to it
  while(!(ADCSRA & _BV(ADIF)))
  ;
  sum=0;
  for (i=0; i<32; i++){
    ADCSRA |= _BV(ADIF); //Clear ADIF by writing 1 to it
    while(!(ADCSRA & _BV(ADIF))) //Wait for conversion complete
    ;
    sum+=ADC;  //ADCL + 256*ADCH
  }
  g.offsetU = sum / 32;
  ADCSRA = _BV(ADEN)| //Enable
          //_BV(ADSC) | //Start conversion
          //_BV(ADATE) | //Free running off
          _BV(ADIE) | //Interrupt ein
          _BV(ADPS2)| //Prescaler=128  
          _BV(ADPS1)|
          _BV(ADPS0);
  ADCSRA |= _BV(ADIF); //Clear ADIF by writing 1 to it


  //Vorlufig: Offset I = Offset U
  //Eigentlich muss bei I=0 auch der Offset des LEM mitgemessen werden.
  //Dazu msste ein Relais rein, welches I unterbricht.
  g.offsetI = g.offsetU;
  
}
//--------------------------------------------------------------------------
void offsetIMessung(void){
  byte i;
  int sum;
  ADMUX = _BV(REFS0)|1; //Internal 5V Ref; Right adjust; MUX=Ch1 (Strom)
  ADCSRA = _BV(ADEN)| //Enable
          _BV(ADSC) | //Start conversion
          _BV(ADATE)| //Auto trigger (free running )
          _BV(ADPS2)| //Prescaler=128  
          _BV(ADPS1)|
          _BV(ADPS0);
  ADCSRB = 0; //free running

  ADCSRA |= _BV(ADIF); //Clear ADIF by writing 1 to it
  while(!(ADCSRA & _BV(ADIF)))
  ;
  sum=0;
  for (i=0; i<32; i++){
    ADCSRA |= _BV(ADIF); //Clear ADIF by writing 1 to it
    while(!(ADCSRA & _BV(ADIF))) //Wait for conversion complete
    ;
    sum+=ADC;  //ADCL + 256*ADCH
  }
  g.offsetI = sum / 32;
  ADCSRA = _BV(ADEN)| //Enable
          //_BV(ADSC) | //Start conversion
          //_BV(ADATE) | //Free running off
          _BV(ADIE) | //Interrupt ein
          _BV(ADPS2)| //Prescaler=128  
          _BV(ADPS1)|
          _BV(ADPS0);
  ADCSRA |= _BV(ADIF); //Clear ADIF by writing 1 to it

}



//--------------------------------------------------------
void delay_ms(int dauer_ms){
#if 1
  //Set up the 8-bit timer 2  (1KHz Generator)
  //------------------------------------------
  TCCR2B = 0; //AUS. 5=Teiler durch 128
  TCCR2A = (1<<WGM21); //CTC _Modus
  OCR2A = 124; //Teilt durch 125

  //Timer luft mit 1 KHz ber 


  //Zhler reseten
  TCNT2 = 0;
  g.ms=dauer_ms;
  g.finished=0;

  //Timer starten
  TIMSK2 = (1<<OCIE2A);
  TCCR2B = 5;  //5= Teiler durch 128

  
  //Warten, bis Timer Endwert erreicht
  while (!g.finished)
  ; 

  
  //Timer 2 abschalten
  TCCR2B = 0;
  TIMSK2 = 0;
  
#else   

//Notlsung ohne Timer. Handkalibriert fr -O0.
int i,j;

  for (i=0; i<dauer_ms; i++){
    for (j=0; j<613; j++){
   
      asm volatile ("push r0");
      asm volatile ("pop r0");
      asm volatile ("NOP");
      asm volatile ("nop");

    }
  }
#endif
}
//--------------------------------------------------------
void USART_Transmit( unsigned char data ){
//--------------------------------------------------------
	/* Wait for empty transmit buffer */
	while ( !( UCSR0A & (1<<UDRE0)) )
	;
	/* Put data into buffer, sends the data */
	UDR0 = data;
}
//--------------------------------------------------------
void sendstringRS232(unsigned char *text){
//--------------------------------------------------------
  while(*text){
    USART_Transmit(*text++);
  }
}


//--------------------------------------------------------
void int3out(U08 pos, int x){
  g.message[pos++]=48+x/100;
  x%=100;
  g.message[pos++]=48+x/10;
  x%=10;
  g.message[pos++]=48+x;
}
//--------------------------------------------------------
void intS3out(U08 pos, int x){
  if (0==x){
    g.message[pos++]=' ';
  }else
  if (x>0){
    g.message[pos++]='L';
  }else{
    g.message[pos++]='C';
    x=0-x;
  }
  g.message[pos++]=48+x/100;
  x%=100;
  g.message[pos++]=48+x/10;
  x%=10;
  g.message[pos++]=48+x;
}
//--------------------------------------------------------

int main (void)
{
int i, loops;
byte *ram;
int x;

long ueff,ieff,wirk;
static byte vollwellenZaehler;
static byte schalter;
static byte ueberlast;
static byte vollwellen;

    cli();
    //buffer = (unsigned char *)&g.istwerteU;

    //----- Disable the watchdog ----------------------------------
    asm("wdr"); //__watchdog_reset();
    /* Clear WDRF in MCUSR */
    MCUSR &= ~(1<<WDRF);
    /* Write logical one to WDCE and WDE */
    /* Keep old prescaler setting to prevent unintentional time-out */
    WDTCSR |= (1<<WDCE) | (1<<WDE);
    /* Turn off WDT */
    WDTCSR = 0x00;


    //---- Set port directions ----------------------------------
    DDRC = 8+16+32;
    PORTC = 0; //alles AUS

    DDRD = 0; //Alles Eingnge (TX wird autom. zum Ausgang)
    PORTD = 255;  //alle Pull-Up Widerstnde ein



    DDRB = 0;
    PORTB = 0; //keine Pull-Ups.

    //----- setup A/D converter  --------------------------------
    //ADC - single Conversion Mode (not Free Running)
    ADMUX = _BV(REFS0); //Internal 5V Ref; Right adjust(10Bit); MUX=Ch0
    ADCSRA = _BV(ADEN)| _BV(ADIE)|_BV(ADPS2)|_BV(ADPS1)|_BV(ADPS0);//Enable; IRQ Enable; Prescaler128 
                                                                   //  gibt bei 16MHz: 125KHz ADC
    //ADC liefert mit knapp 9KHz einen Messwert 

         //Set up the 8-bit timer 0 (500ns Takt)  - Messung von Verweildauer
    TCCR0A = 0;          //Normaler Timer/Counter;
    TCCR0B = _BV(CS01) ; // Prescaler= /8 (gibt 2MHz)
    TCNT0=0;
    //TIFR0 |=1; //Lscht (!) TOV0 Bit

        //Set up the 16-bit timer 1 (winkelmessung).
    TCCR1A =  0; //Normaler Timer/Counter
    TCCR1B =  0;//Spter:  _BV(CS12); // Vorteiler 256
    //Zeitbasis: 62500 KHz 16us 
    //Zhlt max. bis 65535 - also ca. 1 Sekunde
    //Bei L=90 (5ms): Zhlerstand=312.5  180 = 625  // *32=20000 /111 = 180,180180
    //Gradzahl = Zhlerstand /3,466666 
    //Bei L=-0 (20ms): Zhlerstand=1250
    


    //----- RS232 initialisieren --------------------------------------------------

    /* Set baud rate */
    UBRR0H = 0;
    UBRR0L = 103;  //9600 Baud bei 16MHz Takt
    UCSR0A=0; //Clear U2X
  
    /* Enable transmitter, no Interrupts*/
    UCSR0B = (1<<TXEN0);
  
    /* Set frame format: 8data, 1stop bit, no parity */
    UCSR0C = (1<<UCSZ01)|(1<<UCSZ00);


    
    //Globale Variablen mit 0 initialisieren
    ram=(unsigned char *)&g;
    for (i=0; i< sizeof(Tglob); ram[i++]=0); 


    sei(); //Interrupts EIN. Jetzt geht's los!


//while (1){PORTD ^=255 ; delay_ms(1000);};

    g.messen=0;

#if 1
    for (i=7; i>0; i--){
      delay_ms(250);
      loadText(txt03); //"|o|*|u|*|1Wirkleistungsmessgeraet"
      sendstringRS232(g.message);

      delay_ms(100);
      loadText(txt04);//"|3D. Schroeder 01/2009"
      sendstringRS232(g.message);
      delay_ms(100);

    }
#endif

    offsetUMessung();
    offsetIMessung();
    vollwellen=4;
    loops=0;

    //------------- Hauptschleife ------------------------------ 
    while(1){

      //Bereichsschalter auslesen
      schalter = PIND;
      if (!(schalter & 1)){
                         //          10         20 
        offsetIMessung();//0123456789 123456789 1234567
        loadText(txt02);//"|u|*|o|*|2I-Offset=123"
        int3out(19,g.offsetI);
        sendstringRS232(g.message);
        delay_ms(333);
        continue;
      }else  
      if (!(schalter & 4)){
        g.LemWindungen=1;  //400mA 92W
      }else  
      if (!(schalter & 8)){
        g.LemWindungen=4;  //100mA 23W
      }else  
      if (!(schalter & 16)){
        g.LemWindungen=16; //25mA 5,75W
      }else{                      //123456789 123456789 1234567
        loadText(txt01);//"|o|*|u|*|2     <- weiter drehen ->"
        sendstringRS232(g.message);
        delay_ms(333);
        continue;
      }

//      g.verweildauerMax =   0;
//      g.verweildauerMin = 255;
      g.sumSpannung =0;
      g.sumStrom=0;
      g.sumWirk=0;
//      g.sumWinkel=0;
      g.countU=0;
      g.countI=0;
//      g.countPhi=0;
//      g.Umax=0;
//      g.Imax=0;
      vollwellenZaehler = 0;



      //----  Spannungsmessung  ----
      g.spannungsmessung=1;
      //ADC erstmal abschalten, damit er zu definiertem Zeitpunkt gestartet werden kann
      g.messen=0;
      g.nulldurchgaenge=0;
      g.trigger=0;
      ADMUX = _BV(REFS0); //Spannung
//      ADCSRA |= _BV(ADIF); //Clear ADIF by writing 1 to it
      
      //Warten auf Start einer Vollwelle (Nulldurchgang)
      while (!(PINB & 1));
      while (PINB & 1);
      while (!(PINB & 1));
      TCCR1B =  0; //Stopp timer
      TCNT1=0;
      TCCR1B =  _BV(CS12); // Starte Timer mit Vorteiler 256

      //Bin jetzt am Anfang einer Vollwelle
      
      g.messen=1;
//      ADCSRA = _BV(ADEN)| _BV(ADIE)|_BV(ADPS2)|_BV(ADPS1)|_BV(ADPS0);//Enable; IRQ Enable; Prescaler128 
//      ADCSRA |= _BV(ADSC); //Start next conversion



      ADCSRA = _BV(ADEN)| //Enable
              _BV(ADIE) | //Interrupts ein
              _BV(ADSC) | //Start conversion
              _BV(ADATE)| //Auto trigger (free running )
              _BV(ADPS2)| //Prescaler=128  
              _BV(ADPS1)|
              _BV(ADPS0);
      ADCSRB = 0; //free running
      ADCSRA |= _BV(ADIF); //Clear ADIF by writing 1 to it



      
      //Messe bis gengend Vollwellen durchgelaufen
      for (vollwellenZaehler=0; vollwellenZaehler<vollwellen; vollwellenZaehler++){
        while (PINB & 1);
        while (!(PINB & 1));
      };
      //ADC erstmal abschalten, damit er zu definiertem Zeitpunkt gestartet werden kann
      g.messen=0;
      g.nulldurchgaenge=0;
      g.trigger=0;
      ADCSRA = 0;
      
      //----  Strommessung  ----
      g.spannungsmessung=0;
      ADMUX = _BV(REFS0)|1; //Strom
//      ADCSRA |= _BV(ADIF); //Clear ADIF by writing 1 to it
      
      
      //Warten auf Start der nchsten Vollwelle (Nulldurchgang)
      while (PINB & 1);
      while (!(PINB & 1));

      //Bin jetzt am Anfang einer Vollwelle
      TCCR1B =  0; //Stopp timer
      TCNT1=0;
      TCCR1B =  _BV(CS12); // Starte Timer mit Vorteiler 256
      
      g.messen=1;
//      ADCSRA = _BV(ADEN)| _BV(ADIE)|_BV(ADPS2)|_BV(ADPS1)|_BV(ADPS0);//Enable; IRQ Enable; Prescaler128 
//      ADCSRA |= _BV(ADSC); //Start next conversion
      ADCSRA = _BV(ADEN)| //Enable
              _BV(ADIE) | //Interrupts ein
              _BV(ADSC) | //Start conversion
              _BV(ADATE)| //Auto trigger (free running )
              _BV(ADPS2)| //Prescaler=128  
              _BV(ADPS1)|
              _BV(ADPS0);
      ADCSRB = 0; //free running
      ADCSRA |= _BV(ADIF); //Clear ADIF by writing 1 to it

      //Messe bis gengend Vollwellen durchgelaufen
      for (vollwellenZaehler=0; vollwellenZaehler<vollwellen; vollwellenZaehler++){
        while (PINB & 1);
        while (!(PINB & 1));
      };
      g.messen=0;
      ADCSRA = 0;
      
      //---- Auswertung ----
      g.messen=0;
      ueff = (sqrt(g.sumSpannung / g.countU) * U_KALIBRIER)/1000L;  //in V
      ieff = sqrt(g.sumStrom / g.countI);
      if (ieff<2) ieff=0; //Rauschen unterdrcken
      ieff = (ieff* I_KALIBRIER)/100L;   //in 1/10 mA 
      if (ieff > 4000){
        ueberlast=1;
      }else{
        ueberlast=0;
      }
      ieff /= g.LemWindungen;
      wirk = (g.sumWirk / g.countI   * I_KALIBRIER)/1000L * U_KALIBRIER/1000L; 
      wirk /= g.LemWindungen; //in mW
//      wirk /=1000;
//      phi  = g.sumWinkel / g.countPhi;

//      phi=(phi*1000)/PHI_KALIBRIER;
//      if (phi > 180) phi -= 360;

//      sprintf(buffer,"|1|*Ueff=%3.0fV  Ieff=%#5.1fmA|2P=%6.3fW phi=%4ld ",ueff,ieff,wirk,phi); 
//      if (g.countPhi){
//        sprintf(buffer,"|1|*P=%#7.3fW phi=%4ld |2Ueff=%3.0fV  Ieff=%#5.1fmA",wirk,phi,ueff,ieff); 
//      }else{

      loadText(txt05);
       //          10        20        30
       //0123456789 123456789 123456789 123456789
      // |1|*P=123.567W|2Ueff=123V Ieff=123.5mA

      if (wirk<0){
        g.message[5]='-';
        wirk=0-wirk;
      }

      //Leistung
      x = (int)(wirk/1000L);
      int3out(6,x);

      x = (int)(wirk % 1000);
      int3out(10,x);

      //Fhrende Nullen entfernen
      if (48==g.message[6]) g.message[6]=32;
      if ((48==g.message[7]) && (32==g.message[6])) g.message[7]=32;

      //Spannung
      x = ueff;
      int3out(21,x);

      //Strom
      x = (int)(ieff/10);
      int3out(31,x);
      x = ieff%10;
      g.message[35]=48+x;
      //Fhrende Nullen entfernen
      if (48==g.message[31]) g.message[31]=32;
      if ((48==g.message[32]) && (32==g.message[31])) g.message[32]=32;


      sendstringRS232(g.message);

      loadText(txt06);//|3|*Bereich "
      sendstringRS232(g.message);

      switch(g.LemWindungen){
        case 1: 
          loadText(txt07);// 25  100 [400] 
          break;
        case 4:
          loadText(txt08);// 25 [100] 400
          break;
        case 16:
          loadText(txt09);//[25] 100  400
          break;
      }
      sendstringRS232(g.message);
      if (g.error){
        sendstringRS232("?");
      }

      if (ueberlast){
                        //123456789 123456789 1234567
                        // |Gyx= GotoYX x:1..4  y:@..Z
                        //@ABCDEFGHIJKLMNOPQRSTUVWXYZ   
        loadText(txt10);// |G4HUeberlast
        sendstringRS232(g.message);
      }else{
//        sprintf(buffer,"|4Nphi=%d Tmin=%d Ni=%d ",g.countPhi, g.verweildauerMin, g.countI); 
//        sprintf(buffer,"|4Ni=%d Nphi=%d Ioffs=%d", g.countI, g.countPhi, g.offsetI); 
//        sprintf(buffer,"|4Ni=%d Ioffs=%d", g.countI, g.offsetI); 
        
                        // 0123456789 123456789 1234567
#if 0
        loadText(txt11);// |4Ni=123 Nu=123 Ioffs=123
        int3out(5,g.countI);
        int3out(12,g.countU);
        int3out(22,g.offsetI);
        sendstringRS232(g.message);
#else
                        // 0123456789 123456789 123456
        if (g.nulldurchgaenge){                            
          loadText(txt13);// |4Phi=+123 +456 +789 +123 
          intS3out(6,g.phi[0]*17/59);
          intS3out(11,g.phi[1]*17/59);
          intS3out(16,g.phi[2]*17/59);
          intS3out(21,g.phi[3]*17/59);
          sendstringRS232(g.message);
        }
#endif
      }
      loadText(txt12);
      // 0123456789 123456789 123456
      // |G1SNnull=12
      if (g.nulldurchgaenge >9){
        g.message[10]= 48 + (g.nulldurchgaenge / 10);
        g.message[11]= 48 + (g.nulldurchgaenge % 10);
      }else{
        g.message[10]= 48+g.nulldurchgaenge;
        g.message[11]= 32;
      }
      sendstringRS232(g.message);

      g.phi[0]=0;
      g.phi[1]=0;
      g.phi[2]=0;
      g.phi[3]=0;

      //etwas warten, ausser der Schalter wird gedreht
      for (i=0; i<7; i++){
        if (schalter != PIND) break;
        delay_ms(40);
      }

    }//while

    return 0;
}
